# coding: utf-8
import re
import time
# import prompts

from prompts.psyqa.planner import *
from prompts.psyqa.solver import *
from prompts.psyqa.prompt import *

from prompts.tutoring.cima.prompt import *
from prompts.tutoring.cima.planner import *
from prompts.tutoring.cima.solver import *

from run_strategy.dataset import *
from run_strategy.evaluate import *

from tools.retriever import *
from tools.utils import *

from app.sample_resp_cralwer import SampleRespCrawler
from app.chatgpt_resp_cralwer import ChatgptRespCrawler
from app.huggingface_resp_crawler import HuggingFaceRespCrawlerBase

class ReWOO:
    def __init__(self, key_path, dataset, model_type="chatgpt", model_path="", language="", setting="zero-shot",
            temperature=0.7, persona="", top_p=0.95):
        
        self.key_path = key_path
        self.dataset = dataset
        self.model_type = model_type
        self.model_path = model_path
        self.language = language
        self.token_unit_price = get_token_unit_price(model_type)

        self.setting = setting
        if self.dataset == "hotpotqa":
            data_paths = ["./dataset/hotpotQA/hotpot_dev_distractor_v1.json"]
            demo_path = "./cot_retrieval/config/prompt_en.json"
            self.prompt_constructor = HotpotQA(data_paths, demo_path=demo_path, mode="dev")
        elif self.dataset == "psyqa":
            data_paths = ["./dataset/PsyQA/psyqa_test.json"]
            self.prompt_constructor = PsyQA(data_paths, mode="dev")
            self.language = "cn"
        elif self.dataset == "cima":
            data_paths = ["./dataset/tutoring/CIMA_test.json"]
            self.prompt_constructor = StrategyTutoring(data_paths, mode="dev")
            self.language = "en"

        self.retriever = Retriever('bm25')
        self.call_retrieval_times, self.persona_right_count, self.knowledge_right_count = 0, 0, 0
        self.init_model_type(temperature, persona, top_p, model_type)

    def init_model_type(self, temperature, persona, top_p, model):
        if self.model_type in ["gpt-3.5-turbo", "gpt-3.5-turbo-0613", "text-davinci-003"]:
            self.sample_crawler = ChatgptRespCrawler(self.key_path, temperature, persona=persona, top_p=top_p, model=model)
        else:
            self.sample_crawler = HuggingFaceRespCrawlerBase(self.model_type, self.model_path, top_p=top_p, temperature=temperature)

    def plan(self, prompt):
        plan = self.sample_crawler.call_openai_each(prompt)
        total_tokens, plan = get_response_according_to_model_type(plan, self.model_type)

        self.cache["prices"] += total_tokens * self.token_unit_price
        return plan

    def solve(self, prompt):
        generations = self.sample_crawler.call_openai_each(prompt)
        total_tokens, generations = get_response_according_to_model_type(generations, self.model_type)

        self.cache["prices"] += total_tokens * self.token_unit_price
        return generations
    
    def _parse_plans(self, response):
        plans = []
        for line in response.splitlines():
            if line.startswith("Plan:"):
                plans.append(line)
        return plans
    
    def _parse_planner_evidences(self, response):
        evidences = {}
        for line in response.splitlines():
            if line.startswith("#") and line[1] == "E" and line[2].isdigit():
                e, tool_call = line.split(":", 1)
                e, tool_call = e.strip(), tool_call.strip()
                if len(e) == 3:
                    evidences[e] = tool_call
                else:
                    evidences[e] = "No evidence found"
        return evidences
    
    # input: the question line. e.g. "Question: What is the capital of France?"
    def predict_responses(self):
        result = {}
        
        # init the evaluation metrics
        count, rl_scores, f1_scores, avg_bleu_scores = 0, [], [], []
        d1_scores, d2_scores = [], []
        prices, evaluation_elo = [], []
        bleu_scores, bert_scores = [], []

        for sample in self.prompt_constructor.examples[:200]:
            count += 1 # the number of current results

            # init cache
            self.cache = {}
            self.cache["prices"] = 0
            if self.dataset == "psyqa":
                question = sample["question"]
                desc = sample["desc"]
                self.cache["question"] = sample["question"]
                self.cache["description"] = sample["desc"]
                self.cache["response"] = sample["answer"]
            elif self.dataset == "cima":
                self.cache["context"] = sample["context"]
                self.cache["response"] = sample["references"]
                context = sample["context"]

            # run is stateless, so we need to reset the evidences
            self._reinitialize()

            # plan
            if self.dataset == "psyqa":
                planner_prompt = DEFAULT_PREFIX + psyqa_generate_worker_prompt() + DEFAULT_FEWSHOT + psyqa_rewoo_plan_prompt + f"Question: {question}\nDescription: {desc}" + '\n'
            elif self.dataset == "cima":
                planner_prompt = DEFAULT_PREFIX + cima_generate_worker_prompt() + DEFAULT_FEWSHOT + cima_rewoo_plan_prompt + f"Dialogue: {context}" + '\n'

            plan = self.plan(planner_prompt)
            self.plans = self._parse_plans(plan)
            self.planner_evidences = self._parse_planner_evidences(plan)
            # assert len(self.plans) == len(self.planner_evidences)

            # Work
            self._get_worker_evidences()
            worker_log = ""
            for i in range(min(len(self.plans), len(self.worker_evidences))):
                e = f"#E{i + 1}"
                worker_log += self.worker_evidences[e] + " "

            # Solver
            generations = worker_log

            if self.dataset == "cima":
                bert_score, bleu_score = tutoring_scores(generations, self.cache["response"], self.language)
                bleu_scores.append(bleu_score)
                bert_scores.append(bert_score)
            
            f1_scores.append(f1(generations, self.cache["response"], self.language))
            # rl_scores.append(rl(generations, self.cache["response"]))
            prices.append(self.cache["prices"])
            d1_scores.append(distinct_n_sentence_level(generations, 1, lang=self.language))
            d2_scores.append(distinct_n_sentence_level(generations, 2, lang=self.language))
            # avg_bleu_scores.append(avg_bleu(generations, self.cache["response"], lang="cn"))

            if self.dataset == "psyqa":
                evaluation_elo.append({
                    "question": sample["question"],
                    "desc": sample["desc"],
                    "answer": sample["answer"],
                    "generation": generations
                })
            elif self.dataset == "cima":
                evaluation_elo.append({
                    "context": sample["context"],
                    "plan": self.plans,
                    "generation": generations
                })
        
        # save the final evaluation results
        f1_score, rl_score, avg_bleu_score = np.mean(f1_scores), np.mean(rl_scores), np.mean(avg_bleu_scores)
        d1_score, d2_score = np.mean(d1_scores), np.mean(d2_scores)

        result = {
            "method": "rewoo",
            "model_type": self.model_type,
            "f1_score": f1_score,
            "d1_score": d1_score,
            "d2_score": d2_score,
            "avg_bleu_score": avg_bleu_score,
            "price": sum(prices),
            "bleu_score": np.mean(bleu_scores, dtype=np.float64),
            "bert_score": np.mean(bert_scores, dtype=np.float64)
        }

        with open("./exp_output/" + self.dataset + "/rewoo_result.json", "w", encoding="utf-8") as f:
            json.dump(result, f, indent=4, ensure_ascii=False)

        with open("./response_output/" + self.dataset + "/rewoo_result.json", "w", encoding="utf-8") as f:
            json.dump(evaluation_elo, f, indent=4, ensure_ascii=False)

    # use planner evidences to assign tasks to respective workers.
    def _get_worker_evidences(self):
        for e, tool in self.planner_evidences.items():
            tool = tool.lower()         
            if "approval and reassurance" in tool:
                self.worker_evidences[e] = self.approval_reassurance()
            elif "interpretation" in tool:
                self.worker_evidences[e] = self.interpretation()
            elif "direct guidance" in tool:
                self.worker_evidences[e] = self.direct_guidance()
            elif "information" in tool:
                self.worker_evidences[e] = self.inform()
            elif "restatement" in tool:
                self.worker_evidences[e] = self.restatement()
            elif "self-disclosure" in tool:
                self.worker_evidences[e] = self.disclosure()
            elif "hint" in tool:
                self.worker_evidences[e] = self.hint()
            elif "question" in tool:
                self.worker_evidences[e] = self.question()
            elif "correction" in tool:
                self.worker_evidences[e] = self.correction()
            elif "confirmation" in tool:
                self.worker_evidences[e] = self.confirmation()
            elif "others" in tool:
                if self.dataset == "psyqa":
                    self.worker_evidences[e] = self.others_psyqa()
                elif self.dataset == "cima":
                    self.worker_evidences[e] = self.others_cima()
            else:
                self.worker_evidences[e] = "No evidence found"
    
    def _reinitialize(self):
        self.plans = []
        self.planner_evidences = {}
        self.worker_evidences = {}

    def approval_reassurance(self):
        # init question, desc
        question = self.cache["question"]
        desc = self.cache["description"]
        test_prompt = f"Question: {question}\nDescription: {desc}\nApproval and Reassurance: "
        full_prompt = psyqa_chameleon_ar_prompt + "\n\n" + test_prompt

        ar = self.sample_crawler.call_openai_each(full_prompt, max_tokens=32)
        total_tokens, ar = get_response_according_to_model_type(ar, self.model_type)

        # update the cache
        self.cache["ar:output"] = ar
        self.cache["prices"] += total_tokens * self.token_unit_price
        return ar
    
    def interpretation(self):
        # init question, desc
        question = self.cache["question"]
        desc = self.cache["description"]
        test_prompt = f"Question: {question}\nDescription: {desc}\nInterpretation: "
        full_prompt = psyqa_chameleon_interpretation_prompt + "\n\n" + test_prompt

        interpretation = self.sample_crawler.call_openai_each(full_prompt, max_tokens=32)
        total_tokens, interpretation = get_response_according_to_model_type(interpretation, self.model_type)

        # update the cache
        self.cache["interpretation:output"] = interpretation
        self.cache["prices"] += total_tokens * self.token_unit_price
        return interpretation
    
    def direct_guidance(self):
        # init question, desc
        question = self.cache["question"]
        desc = self.cache["description"]
        test_prompt = f"Question: {question}\nDescription: {desc}\nDirect Guidance: "
        full_prompt = psyqa_chameleon_interpretation_prompt + "\n\n" + test_prompt

        direct_guidance = self.sample_crawler.call_openai_each(full_prompt, max_tokens=32)
        total_tokens, direct_guidance = get_response_according_to_model_type(direct_guidance, self.model_type)

        # update the cache
        self.cache["direct_guidance:output"] = direct_guidance
        self.cache["prices"] += total_tokens * self.token_unit_price
        return direct_guidance

    def inform(self):
        # init question, desc
        question = self.cache["question"]
        desc = self.cache["description"]
        test_prompt = f"Question: {question}\nDescription: {desc}\nInformation: "
        full_prompt = psyqa_chameleon_information_prompt + "\n\n" + test_prompt

        information = self.sample_crawler.call_openai_each(full_prompt, max_tokens=32)
        total_tokens, information = get_response_according_to_model_type(information, self.model_type)

        # update the cache
        self.cache["information:output"] = information
        self.cache["prices"] += total_tokens * self.token_unit_price
        return information
    
    def restatement(self):
        # init question, desc
        question = self.cache["question"]
        desc = self.cache["description"]
        test_prompt = f"Question: {question}\nDescription: {desc}\nInformation: "
        full_prompt = psyqa_chameleon_restatement_prompt + "\n\n" + test_prompt

        restatement = self.sample_crawler.call_openai_each(full_prompt, max_tokens=32)
        total_tokens, restatement = get_response_according_to_model_type(restatement, self.model_type)

        # update the cache
        self.cache["restatement:output"] = restatement
        self.cache["prices"] += total_tokens * self.token_unit_price
        return restatement
    
    def disclosure(self):
        # init question, desc
        question = self.cache["question"]
        desc = self.cache["description"]
        test_prompt = f"Question: {question}\nDescription: {desc}\nSelf-disclosure: "
        full_prompt = psyqa_chameleon_self_disclosure_prompt + "\n\n" + test_prompt

        disclosure = self.sample_crawler.call_openai_each(full_prompt, max_tokens=32)
        total_tokens, disclosure = get_response_according_to_model_type(disclosure, self.model_type)

        # update the cache
        self.cache["disclosure:output"] = disclosure
        self.cache["prices"] += total_tokens * self.token_unit_price
        return disclosure
    
    def others_psyqa(self):
        # init question, desc
        question = self.cache["question"]
        desc = self.cache["description"]
        test_prompt = f"Question: {question}\nDescription: {desc}\nOthers: "
        full_prompt = psyqa_chameleon_others_prompt + "\n\n" + test_prompt

        others = self.sample_crawler.call_openai_each(full_prompt, max_tokens=32)
        total_tokens, others = get_response_according_to_model_type(others, self.model_type)

        # update the cache
        self.cache["others:output"] = others
        self.cache["prices"] += total_tokens * self.token_unit_price
        return others

    def others_cima(self):
        # init question, desc
        context = self.cache["context"]
        test_prompt = f"Dialogue: {context}\nOthers: "
        full_prompt = cima_chameleon_others_prompt + "\n\n" + test_prompt

        others = self.sample_crawler.call_openai_each(full_prompt, max_tokens=32)
        total_tokens, others = get_response_according_to_model_type(others, self.model_type)

        # update the cache
        self.cache["others:output"] = others
        self.cache["prices"] += total_tokens * self.token_unit_price
        return others
    
    def hint(self):
        # init question, desc
        context = self.cache["context"]
        test_prompt = f"Dialogue: {context}\nHint: "
        full_prompt = cima_chameleon_hint_prompt + "\n\n" + test_prompt

        hints = self.sample_crawler.call_openai_each(full_prompt, max_tokens=32)
        total_tokens, hints = get_response_according_to_model_type(hints, self.model_type)

        # update the cache
        self.cache["hint:output"] = hints
        self.cache["prices"] += total_tokens * self.token_unit_price
        return hints

    def question(self):
        # init question, desc
        context = self.cache["context"]
        test_prompt = f"Dialogue: {context}\nQuestion: "
        full_prompt = cima_chameleon_question_prompt + "\n\n" + test_prompt

        question = self.sample_crawler.call_openai_each(full_prompt, max_tokens=32)
        total_tokens, question = get_response_according_to_model_type(question, self.model_type)

        # update the cache
        self.cache["question:output"] = question
        self.cache["prices"] += total_tokens * self.token_unit_price
        return question
    
    def confirmation(self):
        # init question, desc
        context = self.cache["context"]
        test_prompt = f"Dialogue: {context}\nConfirmation: "
        full_prompt = cima_chameleon_confirmation_prompt + "\n\n" + test_prompt

        confirmations = self.sample_crawler.call_openai_each(full_prompt, max_tokens=32)
        total_tokens, confirmations = get_response_according_to_model_type(confirmations, self.model_type)

        # update the cache
        self.cache["confirmations:output"] = confirmations
        self.cache["prices"] += total_tokens * self.token_unit_price
        return confirmations
    
    def correction(self):
        # init question, desc
        context = self.cache["context"]
        test_prompt = f"Dialogue: {context}\nCorrection: "
        full_prompt = cima_chameleon_correction_prompt + "\n\n" + test_prompt

        corrections = self.sample_crawler.call_openai_each(full_prompt, max_tokens=32)
        total_tokens, corrections = get_response_according_to_model_type(corrections, self.model_type)

        # update the cache
        self.cache["corrections:output"] = corrections
        self.cache["prices"] += total_tokens * self.token_unit_price
        return corrections
    
    